# -*- coding: utf-8 -*-
"""
Created on Sun Feb 23 2025

@author: Cord Hockemeyer
"""
from kstpy.helpers import kstFrozenset

xpl_basis = {
       kstFrozenset({"a"}),
       kstFrozenset({"b"}),
       kstFrozenset({"a", "c"}),
       kstFrozenset({"b", "c"}),
       kstFrozenset({"a", "b", "d"})
       }




# -*- coding: utf-8 -*-
"""
Created on Sun Feb 23 2025

@author: Cord Hockemeyer
"""
from kstpy.helpers import kstFrozenset
from kstpy.helpers import domain

def constr(structure):
    """ Compute the smallest knowledge space containing a famly of kstFrozensets

    Parameters
    ----------
    structure: set
        Family of kstFrozensets

    Returns
    -------
    Knowledge space

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> constr(xpl_basis)
    """
        
    space = set({kstFrozenset({}), kstFrozenset(domain(structure))})
    space.union(structure)
    for state in structure:
        new_states = set({})
        for s in space:
            if not ((set({state | s})) <= space):
                new_states.add((state | s))
        space = space | new_states
    return space

def basis(structure):
    """ Determine the basis of a knolwdge space/structure
    
    Parameters
    ----------
    structure: set
        Family of kstFrozensets
    
    Returns
    -------
    Basis

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> s = constr(xpl_basis)
    >>> basis(s)
    """
    b = set({})
    for state in structure:
        h = set(state)
        for i in structure:
            if set(i) < set(state):
                h = h - set(i)
            if h == set({}):
                break
        if len(h) > 0:
            b.add(kstFrozenset(state))
    return b

def surmiserelation(structure):
    """Compute the surmise relation for a knowledge structure
    
    Parameters
    ----------

    structure: set
        Family of kstFrozensets
    
    Returns
    -------

    Corresponding surmise relation
    """
    d = domain(structure)
    sr = set({})
    b = basis(structure)

    for i in d:
        for j in d:
            sr.add((i,j))
    for s in b:
        for i in d:
            for j in d:
                if i in s and not j in s:
                    sr.discard((j,i))
                if j in s and not i in s:
                    sr.discard((i,j))
    return(sr)



# -*- coding: utf-8 -*-

"""

Created on Thu Mar 13 2025



@author: Cord Hockemeyer

"""

from kstpy.helpers import kstFrozenset
from kstpy.helpers import itemname
from kstpy.helpers import domain
from kstpy.helpers import vector2kstFrozenset
from kstpy.helpers import kstFrozenset2vector
import re

def readpatterns(filename):
    """Read a set of response patterns from a file

    Parameters
    ----------

    filename: str

    Returns
    -------
    List of patterns (kstFrozensets)
    """
    f = open(filename, "r")
    content = f.read()
    lines = content.split("\n")
    i = 0
    if (lines[0][0] == "#"):   # SRBT format
        i = 1
    pat = re.search("[01]*", lines[i]).group()
    if ((pat == lines[i]) & (len(pat) > 3)):    # File with just the matrix
        noi = len(lines[i])
        nos = len(lines) - i
    else:     # KST or SRBT format: #items, #states, matrix
        noi = int(lines[i])
        nos = int(lines[i+1])
        i = i+2
    d = list([])
    for q in range(noi):
        d.append(itemname(q))
    p = list([])
    for j in range(nos):
        p.append(vector2kstFrozenset(lines[j+i], d))
    return p

def readstructure(filename):
    """Read a knowledge structure from a file

    Parameters
    ----------
    
    filename: str

    Returns
    -------
    Set of states (kstFrozensets)
    """
    return set(readpatterns(filename))

    
def writekst(filename, x):
    """Write a knowledge structure or a data set to a file
    
    Parameters
    ----------

    filename: str
    x: list or set (of kstFrozensets)

    Currently, we write only classical KST format files.
    """
    f = open(filename, "w")
    d = domain(x)
    f.write(str(len(d))+"\n")
    f.write(str(len(x))+"\n")
    for s in x:
        f.write(kstFrozenset2vector(s, d)+"\n")
    f.close()



# -*- coding: utf-8 -*-
"""
Created on Sun Mar  9 13:03:53 2025

@author: hockemey
"""
import more_itertools as mi

class kstFrozenset(frozenset):
    """
    kstFrozenset is a derivative of the frozenset class.
    
    The main reason for its existance is the print function which
    does not show the class anymore. Set operands (|, &, ^, and -)
    have also been re-defined to produce kstFrozensets.
    """
    def __repr__(self):
        if self == frozenset({}):
            return "{}"
        else:
            return set(self).__repr__()

    def __str__(self):
        if self == frozenset({}):
            return "{}"
        else:
            return set(self).__str__()
    
    def __or__(self, value):
        return kstFrozenset(frozenset(self) | frozenset(value))
    
    def __and__(self, value):
        return kstFrozenset(frozenset(self) & frozenset(value))
    
    def __sub__(self, value):
        return kstFrozenset(frozenset(self) - frozenset(value))
    
    def __xor__(self, value):
        return kstFrozenset(frozenset(self) ^ frozenset(value))
    

def itemname(num):
    """Return an Itemname based on the item number
    """
    if (num < 0):
          raise Exception("Sorry, no numbers below zero")
    elif (num < 26):
         return(chr(num+97))
    elif (num < 676):
         return(chr((num//26)+97)+chr((num%26)+97)) 
    elif (num < 17576):
         return(chr((num//676)+97)+chr((num//26)+97)+chr((num%26)+97)) 
    else:
         raise Exception("So far limited to 17576 items")# -*- coding: utf-8 -*-



def domain(structure):
    """ Determine the domain of a set/list of frozensets
    """
    dom = set({})
    for s in structure:
        dom = dom | s
    return sorted(list(dom))


def powerset(domain):
    return mi.powerset_of_sets(domain)

def vector2kstFrozenset(v, d):
    if (len(v) != len(d)):
        raise Exception("Vector and domain do not match!")
    s = set({})
    for i in range(len(v)):
        if (v[i]!= "0"):
            s.add(d[i])
    return kstFrozenset(s)

def kstFrozenset2vector(s, d):
    if (not(s <= set(d))):
        raise Exception("Set and domain do not match!")
    v = ""
    for i in d:
        if i in s:
            v = v + "1"
        else:
            v = v + "0"
    return v



# -*- coding: utf-8 -*-

from kstpy.helpers import domain
from kstpy.helpers import powerset
import collections as c

def distvec(data, structure):
    """Compute a vector of distances from response patterns to a knowledge structure

    Parameters
    ----------
    data: list
        List of kstFrozensets containing response patterns
    structure: set
        Family of kstFrozensets - the knowledge structure

    Returns
    -------
    Vector of distances between response patterns and knowledge structure
    """

    ddom = domain(data)
    sdom = domain(structure)
    if (ddom != sdom):
        raise Exception("Incompativble domains")

    dvec = []
    for pattern in data:
        di = len(ddom)
        for state in structure:
            d = len(pattern ^ state)
            di = min(di, d)
        dvec.append(di)
    return dvec


def difreq(data, structure):
    """Determine a vector of frequencies of distances between a set of response patterns and a knowledge structure

    Parameters
    ----------
    data: list
        List of kstFrozensets containing response patterns
    structure: set
        Family of kstFrozensets - the knowledge structure

    Returns
    -------
    Vector of distance frequencies
    """

    dvec = distvec(data, structure)
    cnt = c.Counter(dvec)
    di = [cnt[x] for x in sorted(cnt.keys())]
    return di


def di(data, structure):
    """Determine the Discrepancy Index
    
     Parameters
    ----------
    data: list
        List of kstFrozensets containing response patterns
    structure: set
        Family of kstFrozensets - the knowledge structure

    Returns
    -------
    Discrepancy Index
   """
    dvec = distvec(data, structure)
    index = sum(dvecx) / len(dvec)
    return index


def da(data, structure):
    """Determine the Distance Agreement coefficient
    
    Parameters
    ----------
    data: list
        List of kstFrozensets containing response patterns
    structure: set
        Family of kstFrozensets - the knowledge structure

    Returns
    -------
    Distance Agreement coefficient
    """
    ddat = di(data, structure)
    p = kstpy.helpers.powerset(domain(structure))
    dpot = di(p, structure)
    return ddat / dpot